#
# utmgeo.pm, en pakke for konvertering av UTM-koordinater til eller
# fra geografiske koordinater
#
# Mere om UTM (Universal Transverse Mercator) kan du lese om p
#
#   http://www.maptools.com/UsingUTM/quickUTM.html
#
# Laget av [?] og Olav Kvittem
# Objektorienterifisert av Salve J. Nilsen


package utmgeo;

# $utmgeo->new()
# - oppretter en instans av utmgeo-klassen;
#
sub new {
    my( $class ) = shift;

    my( $a )   = 6378137;
    my( $f )   = 1/298.257223563;
    my( $e2 )  = $f*(2-$f);
    my( $e2m ) = $e2/(1-$e2);
    my( $n )   = $f/(2-$f);

    my( $instance ) = {
	'a'   => $a,
	'e2'  => $e2,
	'e2m' => $e2m,
	'n'   => $n
	};

    bless $instance, $class;
    return $instance;
}

# $utmgeo->utmtogeo()
# - konverterer et sett med UTM-koordinater til tilsvarende
#   geografiske.
#
sub utmtogeo {
    my( $self ) = shift;
    my( $X, $Y, $sone ) = @_;

    my( $n ) = $self->{'n'};
    my( $e2m ) = $self->{'e2m'};

    my( $psi, $Wpsi, $Vpsi, $N, $my, $THETApsi, $Spsi, $rho, $ny,
       $phim, $dlambda, $dpsi, $phi, $lambda,$lambda0 );

    $psi = $X/(0.9996 * 6366197.724);
    $Vpsi = sqrt(1+$e2m*(cos($psi)**2));
    $Wpsi = sqrt(1.0-$self->{'e2'}*sin($psi)*sin($psi));
    $N = $self->{'a'}/$Wpsi;
    $my = ($Y-500000)/(0.9996*$N);
    $THETApsi = $psi
	-((3/2)*$n-(31/24)*($n**3)
	   -((15/8)*($n**2)-(435/128)*
	     ($n**4))*cos(2*$psi)+(35/2)*($n**3)*(cos(2*$psi)**2)
	   -(315/64)*($n**4)*(cos($psi*2)**3))*sin(2*$psi);
    $Spsi = ($self->{'a'}/(1+$n))*(1+(1/4)*($n**2)+(1/64)*($n**4))*$THETApsi;
    $rho = ($X/0.9996-$Spsi)/$N;
    $ny = $my*(1-($e2m/6)*(cos($psi)**2)*($my**2));
    $phim = $psi+$rho*(1-($e2m/2)*(cos($psi)**2)*($my**2));
    $dlambda = atan2(0.5*(exp($ny)-exp(-$ny))/cos($phim), 1);
    $dpsi = atan2(cos($dlambda)*sin($phim)/cos($phim), 1)-$psi;
    $phi = $psi+($Vpsi**2)*$dpsi-(3/2)*$e2m*sin($psi)*cos($psi)*($dpsi**2);
    $lambda0 = (3+($sone-31)*6)*3.14159/180;
    $lambda = $dlambda+$lambda0;

    return( $phi, $lambda );
}

# $utmgeo->geotoutm()
# - konverterer et sett med geografiske koordinater til tilsvarende
#   UTM-koordinater.
#
sub geotoutm {
    my( $self ) = shift;
    my( $phi, $lambda, $sone ) = @_;

    my( $n ) = $self->{'n'};
    my( $e2m ) = $self->{'e2m'};

    my( $dlambda, $lambda0, $mid, $ny, $ksi, $Wphi, $N, $dgamma,
       $THETAphi, $Sphi, $dX, $X, $Y );

    $lambda0 = (3+($sone-31)*6)*3.14159/180;
    $dlambda = $lambda - $lambda0;
    $mid = cos($phi)*sin($dlambda);
    $ny = (1/2)*log((1+$mid)/(1-$mid));
    $ksi = atan2(((sin($phi)/cos($phi))/cos($dlambda)),1)-$phi;
    $Wphi = sqrt(1-$self->{'e2'}*sin($phi)**2);
    $N = $self->{'a'}/$Wphi;
    $dgamma = $N*$ny*(1+(($e2m/6)*(cos($phi)**2)*($ny**2)));
    $dX = $N*$ksi*(1+($e2m/2*(cos($phi)**2)*$ny**2));
    $THETAphi = $phi
	-((3/2)*$n-(31/24)*($n**3)
	  -((15/8)*($n**2)-(435/128)*
	    ($n**4))*cos(2*$phi)+(35/2)*($n**3)*(cos(2*$phi)**2)
	  -(315/64)*($n**4)*(cos($phi*2)**3))*sin(2*$phi);
    $Sphi = ($self->{'a'}/(1+$n))*(1+(1/4)*($n**2)+(1/64)*($n**4))*$THETAphi;
    $X = 0.9996*($dX+$Sphi);
    $Y = 0.9996*$dgamma+500000;

    return( $X, $Y );
}

# $utmgeo->utmtoutm()
# - Normaliserer et sett av UTM-koordinater fra en sone over p en
#   annen UTM-sone.
#
sub utmtoutm {
    my( $self ) = shift;
    my( $X, $Y, $IS, $US ) = @_;

    my($phi, $lambda, $Xn, $Yn);

    ($phi,$lambda) = $self->utmtogeo($X,$Y,$IS);
    ($Xn, $Yn) = $self->geotoutm($phi, $lambda, $US);
    return( $Xn, $Yn );
}

1;

